---- Karma system stuff

KARMA = {}

-- ply uid -> karma table for disconnected players who might reconnect
KARMA.RememberedPlayers = {}

-- Convars, more convenient access than GetConVar bla bla
KARMA.cv = {}
KARMA.cv.enabled = CreateConVar("ttt_karma", "1")
KARMA.cv.starting = CreateConVar("ttt_karma_starting", "1000")
KARMA.cv.max = CreateConVar("ttt_karma_max", "1000")
KARMA.cv.ratio = CreateConVar("ttt_karma_ratio", "0.001")
KARMA.cv.killpenalty = CreateConVar("ttt_karma_kill_penalty", "15")
KARMA.cv.roundheal = CreateConVar("ttt_karma_round_increment", "5")
KARMA.cv.clean = CreateConVar("ttt_karma_clean_bonus", "20")
KARMA.cv.tbonus = CreateConVar("ttt_karma_traitorkill_bonus", "40")
KARMA.cv.tratio = CreateConVar("ttt_karma_traitordmg_ratio", "0.0003")
KARMA.cv.debug = CreateConVar("ttt_karma_debugspam", "0")

local config = KARMA.cv

local function IsDebug() return config.debug:GetBool() end

local math = math

function KARMA.InitState()
   SetGlobalBool("ttt_karma", config.enabled:GetBool())
end

function KARMA.IsEnabled()
   return GetGlobalBool("ttt_karma", false)
end

-- Compute penalty for hurting someone a certain amount
function KARMA.GetHurtPenalty(victim_karma, dmg)
   return victim_karma * math.Clamp(dmg * config.ratio:GetFloat(), 0, 1)
end

-- Compute penalty for killing someone
function KARMA.GetKillPenalty(victim_karma)
   -- the kill penalty handled like dealing a bit of damage
   return KARMA.GetHurtPenalty(victim_karma, config.killpenalty:GetFloat())
end

-- Compute reward for hurting a traitor (when innocent yourself)
function KARMA.GetHurtReward(dmg)
   return config.max:GetFloat() * math.Clamp(dmg * config.tratio:GetFloat(), 0, 1)
end

-- Compute reward for killing traitor
function KARMA.GetKillReward()
   return KARMA.GetHurtReward(config.tbonus:GetFloat())
end

function KARMA.GivePenalty(ply, penalty)
   ply:SetLiveKarma(math.max(ply:GetLiveKarma() - penalty, 0))
end

function KARMA.GiveReward(ply, reward)
   ply:SetLiveKarma(math.min(ply:GetLiveKarma() + reward, config.max:GetFloat()))
end

function KARMA.ApplyKarma(ply)
   local df = 1

   -- any karma at 1000 or over guarantees a df of 1, only when it's lower do we
   -- need the penalty curve
   if ply:GetBaseKarma() < 1000 then
      df = 1 + -0.0000025 * ((ply:GetBaseKarma() - 1000)^2)
   end

   ply:SetDamageFactor(math.Clamp(df, 0.1, 1.0))

   if IsDebug() then
      print(Format("%s has karma %f and gets df %f", ply:Nick(), ply:GetBaseKarma(), df))
   end
end

-- Handle karma change due to one player damaging another. Damage must not have
-- been applied to the victim yet, but must have been scaled according to the
-- damage factor of the attacker.
function KARMA.Hurt(attacker, victim, dmg_amount)
   if not IsValid(attacker) or not IsValid(victim) then return end
   if attacker == victim then return end
   if not attacker:IsPlayer() or not victim:IsPlayer() then return end

   -- Ignore excess damage
   local hurt_amount = math.min(victim:Health(), dmg_amount)

   if attacker:GetTraitor() == victim:GetTraitor() then
      local penalty = KARMA.GetHurtPenalty(victim:GetLiveKarma(), hurt_amount)

      KARMA.GivePenalty(attacker, penalty)

      attacker:SetCleanRound(false)

      if IsDebug() then
         print(Format("%s (%f) attacked %s (%f) for %d and gets penalised for %f", attacker:Nick(), attacker:GetLiveKarma(), victim:Nick(), victim:GetLiveKarma(), hurt_amount, penalty))
      end
   elseif (not attacker:GetTraitor()) and victim:GetTraitor() then
      local reward = KARMA.GetHurtReward(hurt_amount)
      KARMA.GiveReward(attacker, reward)

      if IsDebug() then
         print(Format("%s (%f) attacked %s (%f) for %d and gets REWARDED %f", attacker:Nick(), attacker:GetLiveKarma(), victim:Nick(), victim:GetLiveKarma(), hurt_amount, reward))
      end
   end
end


-- Handle karma change due to one player killing another.
function KARMA.Killed(attacker, victim)
   if not IsValid(attacker) or not IsValid(victim) then return end
   if attacker == victim then return end
   if not attacker:IsPlayer() or not victim:IsPlayer() then return end

   if attacker:GetTraitor() == victim:GetTraitor() then
      local penalty = KARMA.GetKillPenalty(victim:GetLiveKarma())

      KARMA.GivePenalty(attacker, penalty)

      attacker:SetCleanRound(false)

      if IsDebug() then
         print(Format("%s (%f) killed %s (%f) and gets penalised for %f", attacker:Nick(), attacker:GetLiveKarma(), victim:Nick(), victim:GetLiveKarma(), penalty))
      end
   elseif (not attacker:GetTraitor()) and victim:GetTraitor() then
      local reward = KARMA.GetKillReward()
      KARMA.GiveReward(attacker, reward)

      if IsDebug() then
         print(Format("%s (%f) killed %s (%f) and gets REWARDED %f", attacker:Nick(), attacker:GetLiveKarma(), victim:Nick(), victim:GetLiveKarma(), reward))
      end
   end
end

-- Handle karma regeneration upon the start of a new round
function KARMA.RoundIncrement()
   local healbonus = config.roundheal:GetFloat()
   local cleanbonus = config.clean:GetFloat()

   for _, ply in pairs(player.GetAll()) do
      KARMA.GiveReward(ply, healbonus + (ply:GetCleanRound() and cleanbonus or 0))

      if IsDebug() then
         print(ply, "gets roundincr", healbonus + (ply:GetCleanRound() and cleanbonus or 0))
      end
   end

   -- player's CleanRound state will be reset by the ply class
end

-- When a new round starts, Live karma becomes Base karma
function KARMA.Rebase()
   for _, ply in pairs(player.GetAll()) do
      if IsDebug() then
         print(ply, "rebased from", ply:GetBaseKarma(), "to", ply:GetLiveKarma())
      end

      ply:SetBaseKarma(ply:GetLiveKarma())
   end
end

-- Apply karma to damage factor for all players
function KARMA.ApplyKarmaAll()
   for _, ply in pairs(player.GetAll()) do
      KARMA.ApplyKarma(ply)
   end
end

function KARMA.NotifyPlayers()
   for _, ply in pairs(player.GetAll()) do
      local text = ""
      local df = ply:GetDamageFactor()
      if df > 0.99 then
         text = Format("YOUR KARMA IS %d, AS A RESULT YOU DEAL FULL DAMAGE THIS ROUND!", math.Round(ply:GetBaseKarma()))
      else
         text = Format("YOUR KARMA IS %d, AS A RESULT DAMAGE YOU DEAL IS REDUCED BY %d%%", math.Round(ply:GetBaseKarma()), math.ceil((1 - df) * 100))
      end

      PlayerMsg(ply, text, false)
   end
end

-- These generic fns will be called at round end and start, so that stuff can
-- easily be moved to a different phase
function KARMA.RoundEnd()
   if KARMA.IsEnabled() then
      KARMA.RoundIncrement()

      -- if karma trend needs to be shown in round report, may want to delay
      -- rebase until start of next round
      KARMA.Rebase()
   end
end

function KARMA.RoundBegin()
   KARMA.InitState()

   if KARMA.IsEnabled() then
      KARMA.ApplyKarmaAll()

      KARMA.NotifyPlayers()
   end
end

function KARMA.Remember(ply)
   KARMA.RememberedPlayers[ply:UniqueID()] = ply:GetLiveKarma()
end

function KARMA.Recall(ply)
   return KARMA.RememberedPlayers[ply:UniqueID()]
end
