
---- Player spawning/dying

local math = math
local table = table
local umsg = umsg
local player = player
local timer = timer
local pairs = pairs

CreateConVar("ttt_server_ragdolls", "1", FCVAR_CHEAT)
CreateConVar("ttt_bots_are_spectators", "0", FCVAR_ARCHIVE)
CreateConVar("ttt_dyingshot", "0")

CreateConVar("ttt_killer_dna_range", "550")
CreateConVar("ttt_killer_dna_basetime", "100")

-- The team the player spawns on depends on the round state
local function SelectTeam()
   if GetRoundState() == ROUND_PREP then
      return TEAM_TERROR
   else
      return TEAM_SPEC
   end
end

-- First spawn on the server
function GM:PlayerInitialSpawn( ply )

   ply:SetTeam(SelectTeam())

   -- Change some gmod defaults
   ply:SetCanZoom(false)
   ply:SetJumpPower(160)
   ply:SetSpeed(false)
   ply:SetCrouchedWalkSpeed(0.3)

   -- Always spawn innocent initially, traitor will be selected later
   ply:ResetStatus()

   -- Start off with clean, full karma
   ply:InitKarma()

   -- We never have weapons here, but this inits our equipment state
   ply:StripAll()

   local rstate = GetRoundState()
   -- We should update the traitor list, if we are not about to send it
   if rstate <= ROUND_PREP then
      SendTraitorList(GetTraitorFilter())
      SendConfirmedTraitors(GetInnocentFilter())
      SendDetectiveList()
   end

   -- Game has started, tell this guy where the round is at
   if rstate != ROUND_WAIT then
      SendRoundState(rstate, ply)
      SendConfirmedTraitors(ply)
      SendDetectiveList(ply)
   end

   -- Handle spec bots
   if ply:IsBot() and GetConVar("ttt_bots_are_spectators"):GetBool() then
      ply:SetTeam(TEAM_SPEC)
      ply:SetForceSpec(true)
   end

   GAMEMODE:CheckPlayerReconnected(ply)
end

function GM:PlayerReconnected(ply)
   if KARMA.IsEnabled() then
      -- grab stored karma
      local stored = KARMA.Recall(ply)
      if tonumber(stored) then
         local k = math.Clamp(stored, 0, KARMA.cv.max:GetFloat())
         ply:SetBaseKarma(k)
         ply:SetLiveKarma(k)
         print(ply, "was remembered by the karma system, and gets a karma of", k)
      end
   end
end


function GM:PlayerSpawn(ply)
   -- Some spawns may be tilted
   ply:ResetViewRoll()

   -- Clear out stuff like whether we ordered guns or what bomb code we used
   ply:ResetRoundFlags()

   -- latejoiner, send him some info
   if GetRoundState() == ROUND_ACTIVE then
      SendRoundState(GetRoundState(), ply)
   end

   -- let the client do things on spawn
   SendUserMessage("plyspawned", ply, ply:IsSpec())

   if ply:IsSpec() then
      ply:StripAll()
      ply:Spectate(OBS_MODE_ROAMING)
      return
   end

   ply:UnSpectate()

   -- ye olde hooks
   self:PlayerLoadout(ply)
   self:PlayerSetModel(ply)

   SCORE:HandleSpawn(ply)
end

function GM:IsSpawnpointSuitable(ply, spwn, force, rigged)
   if not ValidEntity(ply) or not ply:IsTerror() then return true end
   if not rigged and (not ValidEntity(spwn) or not spwn:IsInWorld()) then return false end

   -- spwn is normally an ent, but we sometimes use a vector for jury rigged
   -- positions
   local pos = rigged and spwn or spwn:GetPos()

   if not util.IsInWorld(pos) then return false end

   local blocking = ents.FindInBox(pos + Vector( -16, -16, 0 ), pos + Vector( 16, 16, 64 ))

   for k, p in pairs(blocking) do
      if ValidEntity(p) and p:IsPlayer() and p:IsTerror() and p:Alive() then
         if force then
            p:Kill()
         else
            return false
         end
      end
   end

   return true
end

local SpawnTypes = {"info_player_deathmatch", "info_player_combine",
"info_player_rebel", "info_player_counterterrorist", "info_player_terrorist",
"info_player_axis", "info_player_allies", "gmod_player_start",
"info_player_teamspawn"}

function GetSpawnEnts(shuffled, force_all)
   local tbl = {}
   for k, classname in pairs(SpawnTypes) do
      for _, e in pairs(ents.FindByClass(classname)) do
         if ValidEntity(e) then
            table.insert(tbl, e)
         end
      end
   end

   -- Don't use info_player_start unless absolutely necessary, because eg. TF2
   -- uses it for observer starts that are in places where players cannot really
   -- spawn well. At all.
   if force_all or #tbl == 0 then
      for _, e in pairs(ents.FindByClass("info_player_start")) do
         if ValidEntity(e) then
            table.insert(tbl, e)
         end
      end      
   end

   if shuffled then
      table.Shuffle(tbl)
   end

   return tbl
end

-- Generate points next to and above the spawn that we can test for suitability
local function PointsAroundSpawn(spwn)
   if not ValidEntity(spwn) then return {} end
   local pos = spwn:GetPos()
   
   local w, h = 36, 72 -- bit roomier than player hull

   -- all rigged positions
   -- could be done without typing them out, but would take about as much time
   return {
      pos + Vector( w,  0,  0),
      pos + Vector( 0,  w,  0),
      pos + Vector( w,  w,  0),
      pos + Vector(-w,  0,  0),
      pos + Vector( 0, -w,  0),
      pos + Vector(-w, -w,  0),
      pos + Vector(-w,  w,  0),
      pos + Vector( w, -w,  0),
      pos + Vector( 0,  0,  h) -- just in case we're outside
   };
end

function GM:PlayerSelectSpawn(ply)
   if not IsTableOfEntitiesValid(self.SpawnPoints) then
      
      self.SpawnPoints = GetSpawnEnts(true, false)

      -- One might think that we have to regenerate our spawnpoint
      -- cache. Otherwise, any rigged spawn entities would not get reused, and
      -- MORE new entities would be made instead. In reality, the map cleanup at
      -- round start will remove our rigged spawns, and we'll have to create new
      -- ones anyway.
   end

   local num = table.Count(self.SpawnPoints)
   if num == 0 then
      Error("No spawn entity found!\n")
      return
   end

   -- Just always shuffle, it's not that costly and should help spawn
   -- randomness.
   table.Shuffle(self.SpawnPoints)

   -- Optimistic attempt: assume there are sufficient spawns for all and one is
   -- free
   for k, spwn in pairs(self.SpawnPoints) do
      if self:IsSpawnpointSuitable(ply, spwn, false) then
         return spwn
      end
   end

   -- That did not work, so now look around spawns
   local picked = nil

   for k, spwn in pairs(self.SpawnPoints) do
      picked = spwn -- just to have something if all else fails

      -- See if we can jury rig a spawn near this one
      local rigged = PointsAroundSpawn(spwn)
      for _, rig in pairs(rigged) do
         if self:IsSpawnpointSuitable(ply, rig, false, true) then
            local rig_spwn = ents.Create("info_player_terrorist")
            if ValidEntity(rig_spwn) then
               rig_spwn:SetPos(rig)
               rig_spwn:Spawn()

               ErrorNoHalt("WARNING: Map has too few spawn points, using a rigged spawn for ".. tostring(ply) .. "\n")

               self.HaveRiggedSpawn = true
               return rig_spwn
            end
         end
      end
   end

   -- Last attempt, force one
   for k, spwn in pairs(self.SpawnPoints) do
      if self:IsSpawnpointSuitable(ply, spwn, true) then
         return spwn
      end
   end

   return picked
end


function GM:PlayerSetModel(ply)
   local mdl = GAMEMODE.playermodel or "models/player/phoenix.mdl"
   util.PrecacheModel(mdl)
   ply:SetModel(mdl)
end

local function GiveBasicWeapons(ply)
   ply:Give("weapon_zm_improvised")
   ply:Give("weapon_zm_carry")
   ply:Give("weapon_ttt_unarmed")
end

-- Sometimes, in cramped map locations, giving players weapons fails. A timer
-- calling this function is used to get them the weapons anyway as soon as
-- possible.
local function LateLoadout(id)
   local ply = player.GetByID(id)
   if not ValidEntity(ply) then 
      timer.Destroy("lateloadout" .. id)
      return 
   end

   local num = table.Count(ply:GetWeapons())
   if num == 0 and ply:Team() != TEAM_SPEC then
      GiveBasicWeapons(ply)

      num = table.Count(ply:GetWeapons())
      if num > 0 then
         timer.Destroy("lateloadout" .. id)
      end
   end      
end

function GM:PlayerLoadout( ply )
   if ply:Team() != TEAM_SPEC then
      -- clear out equipment flags
      ply:ResetEquipment()

      GiveBasicWeapons(ply)

      local num = table.Count(ply:GetWeapons())
      if num == 0 then
         Msg("Could not spawn weapons for " .. ply:Nick() .. ", will retry.\n")
         timer.Create("lateloadout" .. ply:EntIndex(), 1, 0, LateLoadout, ply:EntIndex())
      end
   end
end

-- Only active players can use kill cmd
function GM:CanPlayerSuicide(ply)
   return ply:IsTerror()
end

function GM:PlayerSwitchFlashlight(ply, on)
   if not ValidEntity(ply) then return false end

   -- always allow turning off
   return (on == false) or ply:IsTerror()
end

function GM:PlayerSpray(ply)
   if not ValidEntity(ply) or not ply:IsTerror() then
      return true -- block
   end
end

function GM:PlayerUse(ply, ent)
   return ply:IsTerror()
end

function GM:KeyPress(ply, key)
   if not ValidEntity(ply) then return end

   -- Spectator keys
   if ply:IsSpec() and not ply:GetRagdollSpec() then

      if ply.propspec then
         return PropSpecKey(ply, key)
      end

      if key == IN_ATTACK then
         ply:ResetViewRoll()

         -- snap to random guy
         ply:Spectate(OBS_MODE_ROAMING)
         ply:SpectateEntity(nil)

         local alive = util.GetAlivePlayers()

         if #alive < 1 then return end

         local target = table.Random(alive)
         if ValidEntity(target) then
            ply:SetPos(target:EyePos())
         end
      elseif key == IN_ATTACK2 then
         ply:ResetViewRoll()

         -- spectate either the next guy or a random guy in chase
         local target = util.GetNextAlivePlayer(ply:GetObserverTarget())

         if ValidEntity(target) then
            ply:Spectate(ply.spec_mode or OBS_MODE_CHASE)
            ply:SpectateEntity(target)
         end
      elseif key == IN_DUCK then
         ply:ResetViewRoll()

         local pos = ply:GetPos()
         local ang = ply:EyeAngles()

         local target = ply:GetObserverTarget()
         if ValidEntity(target) and target:IsPlayer() then
            pos = target:EyePos()
            ang = target:EyeAngles()
         end

         -- reset
         ply:Spectate(OBS_MODE_ROAMING)
         ply:SpectateEntity(nil)

         ply:SetPos(pos)
         ply:SetEyeAngles(ang)
         return true
      elseif key == IN_JUMP then
         ply:ResetViewRoll()

         -- unfuck if you're on a ladder etc
         if not (ply:GetMoveType() == MOVETYPE_NOCLIP) then
            ply:SetMoveType(MOVETYPE_NOCLIP)
         end
      elseif key == IN_RELOAD then
         local tgt = ply:GetObserverTarget()
         if not IsValid(tgt) or not tgt:IsPlayer() then return end

         if not ply.spec_mode or ply.spec_mode == OBS_MODE_CHASE then
            ply.spec_mode = OBS_MODE_IN_EYE
         elseif ply.spec_mode == OBS_MODE_IN_EYE then
            ply.spec_mode = OBS_MODE_CHASE
         end
         -- roam stays roam

         ply:Spectate(ply.spec_mode)
      end
   end

end


function GM:KeyRelease(ply, key)
   if key == IN_USE and ValidEntity(ply) and ply:IsTerror() then
      -- see if we need to do some custom usekey overriding
      local tr = util.QuickTrace(ply:GetShootPos(), ply:GetAimVector() * 84, ply)
      if tr.Hit and ValidEntity(tr.Entity) then
         if tr.Entity.CanUseKey and tr.Entity.UseOverride then
            local phys = tr.Entity:GetPhysicsObject()
            if ValidEntity(phys) and not phys:HasGameFlag(FVPHYSICS_PLAYER_HELD) then
               tr.Entity:UseOverride(ply)
               return true
            else
               -- do nothing, can't +use held objects
               return true
            end
         elseif tr.Entity.player_ragdoll then
            ShowRagdollSearch(ply, tr.Entity, (ply:KeyDown(IN_WALK) or ply:KeyDownLast(IN_WALK)))
            return true
         end
      end
   end

end

-- Normally all dead players are blocked from IN_USE on the server, meaning we
-- can't let them search bodies. This sucks because searching bodies is
-- fun. Hence on the client we override +use for specs and use this instead.
local function SpecUseKey(ply, cmd, arg)
   if ValidEntity(ply) and ply:IsSpec() then
      -- longer range than normal use
      local tr = util.QuickTrace(ply:GetShootPos(), ply:GetAimVector() * 128, ply)
      if tr.Hit and ValidEntity(tr.Entity) then
         if tr.Entity.player_ragdoll then
            ShowRagdollSearch(ply, tr.Entity)
         else
            PropSpecTarget(ply, tr.Entity)
         end
      end      
   end
end
concommand.Add("ttt_spec_use", SpecUseKey)

function GM:PlayerDisconnected(ply)
   -- Prevent the disconnecter from being in the resends
   if IsValid(ply) then
      ply:SetRole(ROLE_NONE)
   end

   if GetRoundState() != ROUND_PREP then
      -- Keep traitor entindices in sync on traitor clients
      SendTraitorList(GetTraitorFilter(false), nil)

      -- Same for confirmed traitors on innocent clients
      SendConfirmedTraitors(GetInnocentFilter(false))

      SendDetectiveList()
   end


   if KARMA.IsEnabled() then
      KARMA.Remember(ply)
   end
   
   -- let fretta maintain disconnected player table
   self.BaseClass:PlayerDisconnected(ply)
end

---- Death affairs

-- If detective mode, announce when someone's body is found
local bodyfound = CreateConVar("ttt_announce_body_found", "1")

function IdentifyBody(ply, rag)
   if not ply:IsTerror() then return end

   -- simplified case for those who die and get found during prep
   if GetRoundState() == ROUND_PREP then
      rag:SetNWBool("found", true)
      return
   end

   local finder = ply:Nick()
   local nick = rag:GetNWString("nick", "")
   local traitor = (rag.was_role == ROLE_TRAITOR)

   -- Announce body
   if bodyfound:GetBool() and not rag:GetNWBool("found", false) then
      local text = finder .. " FOUND THE BODY OF " .. nick .. "."
      local role = rag.was_role
      if role == ROLE_TRAITOR then
         text = text .. " HE WAS A TRAITOR!"
      elseif role == ROLE_DETECTIVE then
         text = text .. " HE WAS A DETECTIVE."
      else
         text = text .. " HE WAS INNOCENT."
      end

      GameMsg(text)
   end

   -- Register find
   if not rag:GetNWBool("found", false) then
      -- will return either false or a valid ply
      local deadply = player.GetByUniqueID(rag.uqid)
      if deadply then
         deadply:SetNWBool("body_found", true)

         if traitor then
            -- update innocent's list of traitors
            SendConfirmedTraitors(GetInnocentFilter(false))
         end

         SCORE:HandleBodyFound(ply, deadply)
      end

      rag:SetNWBool("found", true)
   end

   -- Handle kill list
   for k, vicid in pairs(rag.kills) do
      -- filter out disconnected
      local vic = player.GetByUniqueID(vicid)
      -- is this an unconfirmed dead?
      if ValidEntity(vic) and not vic:GetNWBool("body_found", false) then
         GameMsg(finder .. " CONFIRMED THE DEATH OF " .. vic:Nick() .. ".")

         -- update scoreboard status
         vic:SetNWBool("body_found", true)

         -- however, do not mark body as found. This lets players find the
         -- body later and get the benefits of that
         --local vicrag = vic.server_ragdoll
         --vicrag:SetNWBool("found", true)
      end
   end
end

-- Send a usermessage to client containing search results
function ShowRagdollSearch(ply, rag, covert, long_range)
   if not ValidEntity(ply) or not ValidEntity(rag) then return end

   local nick = rag:GetNWString("nick")
   local traitor = (rag.was_role == ROLE_TRAITOR)
   local role = rag.was_role
   -- these might be nil
   local eq = rag.equipment or EQUIP_NONE
   local code = rag.bomb_code or ""
   local dmg = rag.dmgtype or DMG_GENERIC
   local wep = rag.dmgwep or ""
   local words = rag.last_words or ""
   local hshot = rag.was_headshot or false
   local dtime = rag.time or 0

   local owner = player.GetByUniqueID(rag.uqid)
   owner = IsValid(owner) and owner:EntIndex() or -1

   -- basic sanity check
   if nick == nil or eq == nil or role == nil then return end

   if DetectiveMode() and not covert then
      IdentifyBody(ply, rag)
   end

   local credits = rag:GetNWInt("credits", 0)
   if ply:IsActiveSpecial() and credits > 0 and (not long_range) then
      local text = "YOU FOUND " .. credits .. ((credits == 1) and " CREDIT" or " CREDITS") .. " ON THE BODY!"
      TraitorMsg(ply, text)
      ply:AddCredits(credits)

      rag:SetNWInt("credits", 0)

      ServerLog(ply:Nick() .. " took " .. credits .. " from the body of " .. nick .. "\n")
      SCORE:HandleCreditFound(ply, nick, credits)
   end

   -- identifier so we know whether a ttt_confirm_death was legit
   ply.search_id = rag:EntIndex() + dtime

   -- time of death relative to current time (saves bits)
   if dtime != 0 then
      dtime = math.Round(CurTime() - dtime)
   end

   -- time of dna sample decay relative to current time
   local stime = 0
   if rag.killer_sample then
      stime = math.max(0, rag.killer_sample.t - CurTime())
   end

   -- build list of people this traitor killed
   local kill_entids = {}
   for k, vicid in pairs(rag.kills) do
      -- also send disconnected players as a marker
      local vic = player.GetByUniqueID(vicid)
      table.insert(kill_entids, IsValid(vic) and vic:EntIndex() or -1)
   end

   local lastid = -1
   if rag.lastid and ply:IsActiveDetective() then
      -- if the person this victim last id'd has since disconnected, send -1 to
      -- indicate this
      lastid = IsValid(rag.lastid.ent) and rag.lastid.ent:EntIndex() or -1
   end

   -- If found by detective, send to all, else just the finder
   local receiver = ply 
   if ply:IsActiveDetective() then receiver = nil end

   -- Send a message with basic info
   umsg.Start("ragsrch", receiver)
   umsg.Short(rag:EntIndex()) -- 2 bytes
   umsg.Short(owner)  -- 2 bytes
   umsg.String(nick)
   umsg.Char(eq)      -- 1 byte
   umsg.Char(role)    -- 1 byte
   umsg.String(code)  -- 3 bytes minimum (assuming codelength 3)
   umsg.Long(dmg)     -- 4 bytes, enum goes high
   umsg.String(wep)   -- 2 bytes(?)
   umsg.Bool(hshot)   -- 1 byte
   umsg.Short(dtime)  -- 2 bytes
   umsg.Short(stime)  -- 2 bytes

   umsg.Char(#kill_entids)  -- 1 byte + (2 * #kills) bytes
   for k, idx in pairs(kill_entids) do
      -- might be possible to use chars here but this is safer
      umsg.Short(idx)
   end

   umsg.Short(lastid)

   -- Who found this, so if we get this from a detective we can decide not to
   -- show a window
   umsg.Short(ply:EntIndex())

   -- Will there be a last words umsg coming up?
   umsg.Bool(words != "") -- 1b
   umsg.End()

   if words != "" then
      -- umsgs only have 128 bytes of room, so if last words is really long we
      -- have to truncate
      if string.len(words) > 127 then
         words = string.sub(words, -127)
      end

      umsg.Start("ragsrch_lw", ply)
      umsg.String(words)
      umsg.End()
   end
end

local function IdentifyCommand(ply, cmd, args)
   if not ValidEntity(ply) then return end
   if not #args == 2 then return end
   
   local eidx = tonumber(args[1])
   local id = tonumber(args[2])
   if not eidx or not id then return end

   if not ply.search_id == eidx + id then
      print(ply, "attempted faulty identify")
      return
   end

   ply.search_id = nil

   local rag = Entity(eidx)
   if ValidEntity(rag) then
      if not rag:GetNWBool("found", false) then
         IdentifyBody(ply, rag)
      else
         ply:ChatPrint("The death of this player has already been confirmed!")
      end
   end
end
concommand.Add("ttt_confirm_death", IdentifyCommand)

local function CreateDeathEffect(ent, marked)
   local pos = ent:GetPos() + Vector(0, 0, 20)

   local jit = 35.0

--   for i=1,3 do
      local jitter = Vector(math.Rand(-jit, jit), math.Rand(-jit, jit), 0)
      util.PaintDown(pos + jitter, "Blood", ent)
--   end

   if marked then
      util.PaintDown(pos, "Cross", ent)
   end
end

-- Returns a sample for use in dna scanner if the kill fits certain constraints,
-- else returns nil
local function GetKillerSample(victim, attacker, dmg)
   -- only guns and melee damage, not explosions
   if not (dmg:IsBulletDamage() or dmg:IsDamageType(DMG_SLASH) or dmg:IsDamageType(DMG_CLUB)) then
      return nil
   end

   if (not IsValid(victim)) or (not IsValid(attacker)) then return end

   local dist = victim:GetPos():Distance(attacker:GetPos())

   if dist > GetConVarNumber("ttt_killer_dna_range") then return nil end

   local sample = {}
   sample.killer = attacker
   sample.victim = victim
   sample.t      = CurTime() + (-1 * (0.019 * dist)^2 + GetConVarNumber("ttt_killer_dna_basetime"))

   return sample
end

-- Creates client or server ragdoll depending on settings
local function MakeRagdoll(ply, attacker, dmginfo)
   if not GetConVar("ttt_server_ragdolls"):GetBool() then
      ply:CreateRagdoll()
      return nil -- signifies we should spectate GetRagdollEntity
   else
      local rag = ents.Create("prop_ragdoll")
      if not ValidEntity(rag) then return nil end

      rag:SetPos(ply:GetPos())
      rag:SetModel(ply:GetModel())
      rag:SetAngles(ply:GetAngles())

      rag:Spawn()
      rag:Activate()

      -- nonsolid to players, but can be picked up and shot
      rag:SetCollisionGroup(COLLISION_GROUP_WEAPON)

      -- flag this ragdoll as being a player's
      rag.player_ragdoll = true
      rag:SetNWString("nick", ply:Nick())
      rag.uqid = ply:UniqueID()
      rag:SetNWBool("found", false)

      -- if someone searches this body they can find info on the ply
      rag.equipment = ply:GetEquipmentItems()
      rag.was_role = ply:GetRole()
      rag.bomb_code = ply.bomb_code
      rag.dmgtype = dmginfo:GetDamageType()

      local wep = util.WeaponFromDamage(dmginfo)
      rag.dmgwep = wep and wep:GetClass() or ""

      rag.was_headshot = ply.was_headshot
      rag.time = CurTime()
      rag.kills = table.Copy(ply.kills)

      rag:SetNWInt("credits", ply:GetCredits())

      rag.killer_sample = GetKillerSample(ply, attacker, dmginfo)

      -- position the bones
      local num = rag:GetPhysicsObjectCount()-1
      local v = ply:GetVelocity() / (num / 2)
      for i=0, num do
         local bone = rag:GetPhysicsObjectNum(i)
         if ValidEntity(bone) then
            local bp, ba = ply:GetBonePosition(rag:TranslatePhysBoneToBone(i))
            bone:SetPos(bp)
            bone:SetAngle(ba)
            -- not sure if this will work:
            bone:SetVelocity(v)

            -- increase their weight
--            bone:SetMass( 40 )
         end
      end

      -- create advanced death effects (knives)
      if ply.effect_fn then
         timer.Simple(0, ply.effect_fn, rag)
--         ply.effect_fn(rag)
      end

      return rag -- we'll be speccing this
   end
end

local deathsounds = {
   Sound("player/death1.wav"),
   Sound("player/death2.wav"),
   Sound("player/death3.wav"),
   Sound("player/death4.wav"),
   Sound("player/death5.wav"),
   Sound("player/death6.wav"),
   Sound("vo/npc/male01/pain07.wav"),
   Sound("vo/npc/male01/pain08.wav"),
   Sound("vo/npc/male01/pain09.wav"),
   Sound("vo/npc/male01/pain04.wav"),
   Sound("vo/npc/Barney/ba_pain06.wav"),
   Sound("vo/npc/Barney/ba_pain07.wav"),
   Sound("vo/npc/Barney/ba_pain09.wav"),
   Sound("vo/npc/Barney/ba_ohshit03.wav"), --heh
   Sound("vo/npc/Barney/ba_no01.wav"),
   Sound("vo/npc/male01/no02.wav"),
   Sound("hostage/hpain/hpain1.wav"),
   Sound("hostage/hpain/hpain2.wav"),
   Sound("hostage/hpain/hpain3.wav"),
   Sound("hostage/hpain/hpain4.wav"),
   Sound("hostage/hpain/hpain5.wav"),
   Sound("hostage/hpain/hpain6.wav")
};


local function PlayDeathSound(victim)
   if not ValidEntity(victim) then return end
   
   WorldSound(table.Random(deathsounds), victim:GetShootPos(), 90, 100)
end

-- See if we should award credits now
local function CheckCreditAward(victim, attacker)
   if not IsValid(victim) then return end

   -- DETECTIVE AWARD
   if IsValid(attacker) and attacker:IsPlayer() and attacker:IsActiveDetective() and victim:IsTraitor() then
      local amt = GetConVarNumber("ttt_det_credits_traitordead") or 1
      for _, ply in pairs(player.GetAll()) do
         if ply:IsActiveDetective() then
            ply:AddCredits(amt)
         end
      end

      TraitorMsg(GetDetectiveFilter(true), "DETECTIVES, YOU HAVE BEEN AWARDED " .. amt .. " EQUIPMENT CREDIT" .. (amt > 1 and "S" or "") .. " FOR YOUR PERFORMANCE.")
   end


   -- TRAITOR AWARD
   if (not victim:IsTraitor()) and (not GAMEMODE.AwardedCredits or GetConVar("ttt_credits_award_repeat"):GetBool()) then
      local inno_alive = 0
      local inno_dead = 0
      for _, ply in pairs(player.GetAll()) do
         if not ply:GetTraitor() then
            if ply:IsTerror() then
               inno_alive = inno_alive + 1
            elseif ply:IsDeadTerror() then
               inno_dead = inno_dead + 1
            end
         end
      end

      -- we check this at the death of an innocent who is still technically
      -- Alive(), so add one to dead count and sub one from living
      inno_dead = inno_dead + 1
      inno_alive = math.max(inno_alive - 1, 0)
      inno_total = inno_dead + inno_alive

      -- Only repeat-award if we have reached the pct again since last time
      if GAMEMODE.AwardedCredits then
         inno_dead = inno_dead - GAMEMODE.AwardedCreditsDead
      end

      local pct = inno_dead / inno_total
      if pct >= GetConVarNumber("ttt_credits_award_pct") then
         -- Traitors have killed sufficient people to get an award
         local amt = GetConVarNumber("ttt_credits_award_size")

         -- If size is 0, awards are off
         if amt > 0 then
            TraitorMsg(GetTraitorFilter(true), "TRAITORS, YOU HAVE BEEN AWARDED " .. amt .. " EQUIPMENT CREDIT" .. (amt > 1 and "S" or "") .. " FOR YOUR PERFORMANCE.")

            for _, ply in pairs(player.GetAll()) do
               if ply:IsActiveTraitor() then
                  ply:AddCredits(amt)
               end
            end
         end

         GAMEMODE.AwardedCredits = true
         GAMEMODE.AwardedCreditsDead = inno_dead + GAMEMODE.AwardedCreditsDead
      end
   end
end

function GM:DoPlayerDeath(ply, attacker, dmginfo)
   if ply:IsSpec() then
      return
   end

   -- Glitch in engine/physics/whatever causes player who stands on dying person
   -- to gain massive fall speed. Workaround places him such that he is no
   -- longer standing on the victim the moment he kills.
   if IsValid(attacker) and attacker:IsPlayer() and ply == attacker:GetGroundEntity() then
      attacker:SetPos(attacker:GetPos() - Vector(0, 0, 1))
   end

   -- Experimental: Fire a last shot if ironsighting and not headshot
   if GetConVar("ttt_dyingshot"):GetBool() then
      local wep = ply:GetActiveWeapon()
      if ValidEntity(wep) and wep.DyingShot and not ply.was_headshot and dmginfo:IsBulletDamage() then
         local fired = wep:DyingShot()
         if fired then
            return
         end
      end

      -- Note that funny things can happen here because we fire a gun while the
      -- player is dead. Specifically, this DoPlayerDeath is run twice for
      -- him. This is ugly, and we have to return the first one to prevent crazy
      -- shit.
   end

   -- Drop all weapons
   for k, wep in pairs(ply:GetWeapons()) do
      DropNotifiedWeapon(ply, wep) -- with ammo in them
      wep:DampenDrop()
   end

   -- Create ragdoll and hook up marking effects
   local rag = MakeRagdoll(ply, attacker, dmginfo)
   ply.server_ragdoll = rag -- nil if clientside
   
   CreateDeathEffect(ply, false)

   util.StartBleeding(rag, dmginfo:GetDamage(), 15)

   -- Score only when there is a round active.
   if GetRoundState() == ROUND_ACTIVE then
      SCORE:HandleKill(ply, attacker, dmginfo)

      if ValidEntity(attacker) and attacker:IsPlayer() then
         attacker:RecordKill(ply)

         DamageLog(string.format("KILL:\t %s [%s] killed %s [%s]\n", attacker:Nick(), (attacker:GetTraitor() and "TRAITOR" or "innocent"), ply:Nick(), (ply:GetTraitor() and "TRAITOR" or "innocent")))
      else
         DamageLog(string.format("KILL:\t <something/world> killed %s [%s]\n", ply:Nick(), (ply:GetTraitor() and "TRAITOR" or "innocent")))
      end
   end

   -- Clear out any weapon or equipment we still have
   ply:StripAll()

   -- Tell the client to send their chat contents
   ply:SendLastWords(dmginfo)

   local killwep = util.WeaponFromDamage(dmginfo)

   -- headshots, knife damage, and weapons tagged as silent all prevent death
   -- sound from occurring
   if not (ply.was_headshot or
           dmginfo:IsDamageType(DMG_SLASH) or
           (IsValid(killwep) and killwep.IsSilent)) then
      PlayDeathSound(ply)
   end

   --- Credits

   CheckCreditAward(ply, attacker)

   -- Check for T killing D or vice versa
   if IsValid(attacker) and attacker:IsPlayer() then
      local reward = 0
      if attacker:IsActiveTraitor() and ply:GetDetective() then
         reward = math.ceil(GetConVarNumber("ttt_credits_detectivekill"))
      elseif attacker:IsActiveDetective() and ply:GetTraitor() then
         reward = math.ceil(GetConVarNumber("ttt_det_credits_traitorkill"))
      end

      if reward > 0 then
         attacker:AddCredits(reward)

         TraitorMsg(attacker, "YOU HAVE RECEIVED " ..  reward .. ((reward == 1) and " CREDIT" or "CREDITS") .. " FOR KILLING A " .. (ply:GetTraitor() and "TRAITOR" or "DETECTIVE") .. ".")
      end
   end
end

function GM:PlayerDeath( victim, infl, attacker)
   -- tell no one
   self:PlayerSilentDeath(victim)

   victim:SetTeam(TEAM_SPEC)
   victim:Freeze(false)
   victim:SetRagdollSpec(true)
   victim:Spectate(OBS_MODE_CHASE)

   local rag_ent = victim.server_ragdoll or victim:GetRagdollEntity()
   victim:SpectateEntity(rag_ent)

   victim:Flashlight(false)

   victim:Extinguish()

   -- Generic message
   if GetConVar("ttt_announce_deaths"):GetBool() and not DetectiveMode() then
      if victim:GetTraitor() then
         GameMsg("A TRAITOR HAS DIED!")
      else
         GameMsg("AN INNOCENT TERRORIST HAS DIED!")
      end
   end

   KARMA.Killed(attacker, victim)

   SendUserMessage("plydied", victim)

   if HasteMode() then
      IncRoundEnd(GetConVar("ttt_haste_minutes_per_death"):GetFloat() * 60)
   end
end

-- kill hl2 beep
function GM:PlayerDeathSound() return true end

function GM:PlayerDeathThink(ply)
   if ply:GetRagdollSpec() then
      if ply:KeyPressed(IN_ATTACK) or CurTime() > ply.spec_ragdoll_start + 5 then
         ply:SetRagdollSpec(false)
         ply:Spectate(OBS_MODE_ROAMING)
      end
   elseif ply:GetMoveType() < MOVETYPE_NOCLIP and ply:GetMoveType() > 0 or ply:GetMoveType() == MOVETYPE_LADDER then
      ply:Spectate(OBS_MODE_ROAMING)
   end
end

function GM:ScalePlayerDamage(ply, hitgroup, dmginfo)
   -- For unknown reasons, damage is doubled after we see it. Compensate for
   -- that here.
   dmginfo:SetDamage(dmginfo:GetBaseDamage() / 2)

   ply.was_headshot = false

   if dmginfo:IsBulletDamage() and ply:HasEquipmentItem(EQUIP_ARMOR) then
      -- Body armor nets you a 20% damage reduction
      dmginfo:ScaleDamage(0.8)
   end

   -- actual damage scaling
   if hitgroup == HITGROUP_HEAD then
      -- headshot if it was dealt by a bullet
      ply.was_headshot = dmginfo:IsBulletDamage()

      local wep = util.WeaponFromDamage(dmginfo)
      local cls = wep and wep:GetClass() or ""
      -- standard multiplier is three times the damage for headshots
      local s = 3

      if cls == "weapon_zm_shotgun" then
         -- scale random headshots made at longer range down
         local att = dmginfo:GetAttacker()
         if not ValidEntity(att) then return end

         local dist = ply:GetPos():Distance(att:GetPos())
         local d = math.max(0, dist - 150)
         
         -- decay from 3.1 to 1 slowly as distance increases past 150 units
         s = 1 + math.max(0, (2.1 - 0.002 * (d ^ 1.25)))

      elseif cls == "weapon_zm_rifle" or cls == "weapon_zm_revolver" then
         -- these weapons have instakill headshots
         s = 4 
      end

      dmginfo:ScaleDamage(s)
   elseif (hitgroup == HITGROUP_LEFTARM or
       hitgroup == HITGROUP_RIGHTARM or 
       hitgroup == HITGROUP_LEFTLEG or
       hitgroup == HITGROUP_LEFTLEG or
       hitgroup == HITGROUP_GEAR ) then
      
      dmginfo:ScaleDamage(0.55)
   end
end

-- The GetFallDamage hook does not get called until around 600 speed, which is a
-- rather high drop already. Hence we do our own fall damage handling in
-- OnPlayerHitGround.
function GM:GetFallDamage(ply, speed)
   return 1
end

local fallsounds = {
   Sound("player/damage1.wav"),
   Sound("player/damage2.wav"),
   Sound("player/damage3.wav")
};

function GM:OnPlayerHitGround(ply, in_water, on_floater, speed)
   if in_water or speed < 425 or not IsValid(ply) then return end

   -- Everything over a threshold hurts you, rising exponentially with speed
   local damage = math.pow(0.05 * (speed - 425), 1.75)

   -- I don't know exactly when on_floater is true, but it's probably when
   -- landing on something that is in water.
   if on_floater then damage = damage / 2 end

   if math.floor(damage) > 0 then
      local dmg = DamageInfo()
      dmg:SetDamageType(DMG_FALL)
      dmg:SetAttacker(GetWorldEntity())
      dmg:SetInflictor(GetWorldEntity())
      dmg:SetDamageForce(Vector(0,0,1))
      dmg:SetDamage(damage)

      if ply.was_pushed and ValidEntity(ply.was_pushed.att) then
         -- was pushed recently
         if ply.was_pushed.t and ply.was_pushed.t > CurTime() - 5 then
            dmg:SetAttacker(ply.was_pushed.att)
         end
      end

--      print("Dealing fall damage:", damage, "for speed", speed)
      ply:TakeDamageInfo(dmg)

      -- play CS:S fall sound if we got somewhat significant damage
      if damage > 5 then
         WorldSound(table.Random(fallsounds), ply:GetShootPos(), 55 + math.Clamp(damage, 0, 50), 100)
      end
   end
end

local ttt_postdm = CreateConVar("ttt_postround_dm", "0", FCVAR_NOTIFY)

-- No damage during prep, etc
function GM:EntityTakeDamage(ent, infl, att, amount, dmginfo)
   if not ValidEntity(ent) then return end

   local rs = GetRoundState()

   if rs == ROUND_PREP or (rs == ROUND_POST and not ttt_postdm:GetBool()) then
      -- if player vs player damage, or if damage versus a prop, then zero
      if (ent:IsExplosive() or (ent:IsPlayer() and ValidEntity(att) and att:IsPlayer())) then
         dmginfo:ScaleDamage(0)
         dmginfo:SetDamage(0)
      end
   elseif ent:IsPlayer() then
      if IsValid(att) and ent != att and att:IsPlayer() and rs == ROUND_ACTIVE then
         DamageLog(string.format("DMG:\t %s [%s] damaged %s [%s] for %d dmg\n", att:Nick(), (att:GetTraitor() and "TRAITOR" or "innocent"), ent:Nick(), (ent:GetTraitor() and "TRAITOR" or "innocent"), dmginfo:GetDamage()))
      end

      -- scale phys damage
      if dmginfo:IsDamageType(DMG_CRUSH) then
         dmginfo:ScaleDamage(0.25)
         if ValidEntity(infl) and ValidEntity(infl:GetOwner()) and infl:GetOwner():IsPlayer() then
            dmginfo:ScaleDamage(0)
            dmginfo:SetDamage(0)
         end
      end

      -- handle fire attacker
      if ent.ignite_info and dmginfo:IsDamageType(DMG_DIRECT) then
         local datt = dmginfo:GetAttacker()
         if not ValidEntity(datt) or not datt:IsPlayer() then
            if ValidEntity(ent.ignite_info.att) and ValidEntity(ent.ignite_info.infl)then
               dmginfo:SetAttacker(ent.ignite_info.att)
               dmginfo:SetInflictor(ent.ignite_info.infl)
            end
         end
      end

      -- start painting blood decals
      util.StartBleeding(ent, dmginfo:GetDamage(), 5)

      -- do karma stuff with pvp damage
      if ent != att and IsValid(att) and att:IsPlayer() and rs == ROUND_ACTIVE and dmginfo:GetDamage() > 0 then
         --print("Karma dmg pre", dmginfo:GetDamage(), att:GetDamageFactor())

         -- scale everything to damage factor except the knife, because it
         -- assumes a kill
         if not dmginfo:IsDamageType(DMG_SLASH) then
            dmginfo:ScaleDamage(att:GetDamageFactor())
         end

         --print("Karma dmg post", dmginfo:GetDamage())
         KARMA.Hurt(att, ent, dmginfo:GetDamage())
      end
   end

   --Msg("TD: " .. tostring(dmginfo:GetDamage()) .. "\n")
end

-- Drowning and such
function GM:Tick()
   -- three cheers for micro-optimizations
   local tm = nil
   local ply = nil
   local plys = player.GetAll()
   for i= 1, #plys do
      ply = plys[i]
      tm = ply:Team()
      if tm == TEAM_TERROR and ply:Alive() then
         -- Drowning
         if ply:WaterLevel() == 3 then
            if ply.drowning then
               if ply.drowning < CurTime() then
                  local dmginfo = DamageInfo()
                  dmginfo:SetDamage(10)
                  dmginfo:SetDamageType(DMG_DROWN)
                  dmginfo:SetAttacker(GetWorldEntity())
                  dmginfo:SetInflictor(GetWorldEntity())
                  dmginfo:SetDamageForce(Vector(0,0,1))

                  ply:TakeDamageInfo(dmginfo)

                  -- have started drowning properly
                  ply.drowning = CurTime() + 1
               end
            else
               -- will start drowning soon
               ply.drowning = CurTime() + 10
            end
         else
            ply.drowning = nil
         end

         -- Slow down ironsighters
         local wep = ply:GetActiveWeapon()
         if ValidEntity(wep) and wep:GetIronsights() then
            ply:SetSpeed(true)
         else
            ply:SetSpeed(false)
         end
      elseif tm == TEAM_SPEC and ply.propspec then
         PropSpecRecharge(ply)

         if IsValid(ply:GetObserverTarget()) then
            ply:SetPos(ply:GetObserverTarget():GetPos())
         end
      end
   end
end

function GM:ShowHelp(ply)
   if ValidEntity(ply) then
      ply:ConCommand("ttt_helpscreen")
   end
end

function GM:PlayerRequestTeam(ply, teamid)
end

-- Implementing stuff that should already be in gmod, chpt. 389
function GM:PlayerEnteredVehicle(ply, vehicle, role)
   if IsValid(vehicle) then
      vehicle:SetNWEntity("ttt_driver", ply)
   end
end

function GM:PlayerLeaveVehicle(ply, vehicle)
   if IsValid(vehicle) then
      -- setting nil will not do anything, so bogusify
      vehicle:SetNWEntity("ttt_driver", vehicle)
   end
end
