
-- Pool all SWEP classnames, as we will be sending some over the wire regularly
for k, w in pairs(weapons.GetList()) do
   if w then
      umsg.PoolString(w.Classname)
   end
end

---- Weapon system, pickup limits, etc

function TypeForWeapon(class)
   local tbl = util.WeaponForClass(class)
   return tbl and tbl.Kind or WEAPON_NONE
end

-- Define what weapons are allowed given a set of carried weapons
function CanCarryWeapon(wep, carried)
   if not wep then return false end

   local weptype = wep.Kind
   for k, v in pairs(carried) do
      if v.Kind == weptype then
--         Msg(v:GetClass() .. " is the same type as " .. wep .. "\n")
         return false
      end
   end

   return true
end

-- Prevent players from picking up multiple weapons of the same type etc
function GM:PlayerCanPickupWeapon(ply, wep)
   if not ValidEntity(wep) and not ValidEntity(ply) then return end

   -- Disallow picking up for ammo
   if ply:HasWeapon(wep:GetClass()) then
      return false
   elseif not CanCarryWeapon(wep, ply:GetWeapons()) then
      return false
   elseif (wep.Kind == WEAPON_EQUIP) and wep.IsDropped and not ply:KeyDown(IN_USE) then
      return false
   end

   local tr = util.TraceEntity({start=wep:GetPos(), endpos=ply:GetShootPos(), mask=MASK_SOLID}, wep)
--   local tr = util.TraceLine({start=wep:GetPos(), endpos=ply:GetShootPos(), mask=MASK_SOLID})
   if tr.Fraction == 1.0 or tr.Entity == ply then
      wep:SetPos(ply:GetShootPos())
   end

   return true
end

---- Weapon switching
function ForceWeaponSwitch(ply, cmd, args)
   -- Turns out even SelectWeapon refuses to switch to empty guns, gah.
   -- Worked around it by giving every weapon a single Clip2 round.
   -- Works because no weapon uses those.
   local wepname = args[1]
   local wep = ply:GetWeapon(wepname)
   if ValidEntity(wep) then
      -- Weapons apparently not guaranteed to have this
      if wep.SetClip2 then
         wep:SetClip2(1)
      end
      ply:SelectWeapon(wepname)
   end
end

concommand.Add("wepswitch", ForceWeaponSwitch)

---- Weapon dropping

function DropNotifiedWeapon(ply, wep)
   if ValidEntity(ply) and ValidEntity(wep) then
      -- Hack to tell the weapon it's about to be dropped and should do what it
      -- must right now
      if wep.PreDrop then
         wep:PreDrop()
      end

      -- Tag this weapon as dropped, so that if it's a special weapon we do not
      -- auto-pickup when nearby.
      wep.IsDropped = true

      ply:DropWeapon(wep)

      wep:PhysWake()

      -- After dropping a weapon, always switch to holstered, so that traitors
      -- will never accidentally pull out a traitor weapon
      ply:SelectWeapon("weapon_ttt_unarmed")
   end
end

function DropActiveWeapon(ply)
   if not ValidEntity(ply) then return end

   local wep = ply:GetActiveWeapon()

   if not ValidEntity(wep) then return end

   if wep.AllowDrop == false then
      return
   end

   local tr = util.QuickTrace(ply:GetShootPos(), ply:GetAimVector() * 32, ply)

   if tr.HitWorld then
      ply:ChatPrint("You have no room here to drop your weapon!")
      return
   end

   DropNotifiedWeapon(ply, wep)
end

concommand.Add("ttt_dropweapon", DropActiveWeapon)

-- Give a weapon to a player. If the initial attempt fails due to heisenbugs in
-- the map, keep trying until the player has moved to a better spot where it
-- does work.
local function GiveEquipmentWeapon(uid, cls)
   -- Referring to players by UID because a player may disconnect while his
   -- unique timer still runs, in which case we want to be able to stop it. For
   -- that we need its name, and hence his uid.
   local ply = player.GetByUniqueID(uid)
   local tmr = "give_equipment" .. tostring(uid)

   if (not ValidEntity(ply)) or (not ply:IsActiveSpecial()) then
      timer.Destroy(tmr)
      return
   end

   -- giving attempt, will fail if we're in a crazy spot in the map or perhaps
   -- other glitchy cases
   local w = ply:Give(cls)

   if (not IsValid(w)) or (not ply:HasWeapon(cls)) then
      if not timer.IsTimer(tmr) then
         timer.Create(tmr, 1, 0, GiveEquipmentWeapon, uid, cls)
      end

      -- we will be retrying
   else
      -- can stop retrying, if we were
      timer.Destroy(tmr)

      if w.WasBought then
         -- some weapons give extra ammo after being bought, etc
         w:WasBought(ply)
      end
   end
end

-- List of equipment ids traitors are allowed to buy. Necessary to make sure
-- they don't buy detective-only stuff and vice versa.
-- Previously both items and weapons had to be in here. Now it's only items.
local traitor_whitelist = {
   -- items
   EQUIP_ARMOR,
   EQUIP_RADAR,
   EQUIP_DISGUISE,

   -- weapons now define this themselves in their CanBuy table
};

local detective_whitelist = {
   -- item
   EQUIP_RADAR,
--   EQUIP_FOOTSTEPS,
};

function OrderEquipment(ply, cmd, args)
   if not ValidEntity(ply) or #args != 1 then return end

   if not (ply:IsActiveTraitor() or ply:IsActiveDetective()) then return end

   if ply:GetCredits() < 1 then
      TraitorMsg(ply, "YOU HAVE NO CREDITS TO SPEND ON EQUIPMENT!")
      return
   end

   -- it's an item if the arg is an id instead of an ent name
   local id = args[1]
   local is_item = tonumber(id)

   -- we use weapons.GetStored to save time on an unnecessary copy, we will not
   -- be modifying it
   local swep_table = (not is_item) and weapons.GetStored(id) or nil

   -- some weapons can only be bought once per player per round, this used to be
   -- defined in a table here, but is now in the SWEP's table
   if swep_table and swep_table.LimitedStock and ply:HasBought(id) then
      TraitorMsg(ply, "THIS WEAPON IS OUT OF STOCK: YOU ALREADY BOUGHT IT EARLIER THIS ROUND!")
      return
   end

   local received = false

   if is_item then
      -- item whitelist check
      local whitelist = (ply:GetTraitor() and traitor_whitelist) or (ply:GetDetective() and detective_whitelist) or nil

      if not whitelist then return end

      if not table.HasValue(whitelist, is_item or id) then
         TraitorMsg(ply, "YOU ARE NOT ALLOWED TO BUY THIS ITEM.")
         print(ply, "tried to buy item not buyable for his class.")
         return
      end

      -- ownership check and finalise
      id = tonumber(id)
      if id and EQUIP_NONE < id and id < EQUIP_MAX then
         if not ply:HasEquipmentItem(id) then
            ply:GiveEquipmentItem(id)
            received = true

            if id == EQUIP_RADAR then
               -- wait until the client knows we have a radar
               timer.Simple(0.5, function()
                                    if IsValid(ply) then
                                       ply:ConCommand("ttt_radar_scan")
                                    end
                                 end)
            end
         else
            TraitorMsg(ply, "YOU ALREADY OWN THIS EQUIPMENT ITEM!")
         end
      end
   elseif swep_table then
      -- weapon whitelist check
      if not table.HasValue(swep_table.CanBuy, ply:GetRole()) then
         TraitorMsg(ply, "YOU ARE NOT ALLOWED TO BUY THIS.")
         print(ply, "tried to buy weapon his role is not permitted to buy")
         return
      end

      -- no longer restricted to only WEAPON_EQUIP weapons, just anything that
      -- is whitelisted and carryable
      if CanCarryWeapon(swep_table, ply:GetWeapons()) then
         GiveEquipmentWeapon(ply:UniqueID(), id)

         received = true
      else
         TraitorMsg(ply, "YOU ARE ALREADY CARRYING A WEAPON OF THIS TYPE!")
      end
   end

   if received then
      ply:SubtractCredits(1)
      TraitorMsg(ply, "YOU HAVE RECEIVED YOUR SPECIAL EQUIPMENT.")

      ply:AddBought(id)
   end
end
concommand.Add("ttt_order_equipment", OrderEquipment)

-- Destroy any equipment we have foreverrrr
local function DestroyEquipment(ply, cmd, args)
   if not ValidEntity(ply) or not ply:IsActiveTraitor() then return end

   if not ply:HasEquipment() then return end

   if ply:HasEquipmentItem() then
      ply:SetEquipmentItem(EQUIP_NONE)
   else
      for _, wep in pairs(ply:GetWeapons()) do
         if ValidEntity(wep) and wep.Kind == WEAPON_EQUIP then
            wep:Remove()
         end
      end
   end
   
   TraitorMsg(ply, "ALL YOUR SPECIAL EQUIPMENT HAS BEEN DESTROYED.")
end
concommand.Add("ttt_destroy_equipment", DestroyEquipment)

local function SetDisguise(ply, cmd, args)
   if not ValidEntity(ply) or not ply:IsActiveTraitor() then return end

   if ply:HasEquipmentItem(EQUIP_DISGUISE) then
      local state = #args == 1 and tobool(args[1])

      ply:SetNWBool("disguised", state)
      ply:ChatPrint("Disguise " .. (state and "enabled!" or "disabled."))
   else
      ply:ChatPrint("You do not have a disguise!")      
   end
end
concommand.Add("ttt_set_disguise", SetDisguise)

local function CheatCredits(ply)
   if server_settings.Bool("sv_cheats", false) and ValidEntity(ply) then
      ply:AddCredits(10)
   end
end
concommand.Add("ttt_cheat_credits", CheatCredits)

-- Protect against non-TTT weapons that may break the HUD
function GM:WeaponEquip(wep)
   if ValidEntity(wep) then
      local cls = string.sub(wep:GetClass(), 8, 9)
      if cls != "zm" and cls != "tt" then
         -- only remove if they lack critical stuff
         if not wep.Kind then
            Warning("Equipped weapon " .. wep:GetClass() .. " is not compatible with TTT\n")
            wep:Remove()
         end
      end
   end
end


